"""Interface for running an exported NengoEdge model in TFLite format."""

import warnings
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import numpy as np

from nengo_edge import config

try:
    import tensorflow as tf
except ImportError:  # pragma: no cover
    warnings.warn("TensorFlow is not installed; cannot use TFLiteRunner.")


class TFLiteRunner:
    """
    Run an exported TFLite model.

    Parameters
    ----------
    directory : Union[str, Path]
        Path to the directory containing the exported model files.
    """

    def __init__(self, directory: Union[str, Path]):
        self.directory = Path(directory)
        self.model_params, self.preprocessing = config.load_params(self.directory)[:2]

        # Build interpreters
        model_path = self.directory / "model.tflite"
        if not model_path.exists():
            raise FileNotFoundError(f"Could not find model file ({model_path})")
        self.model_interpreter = tf.lite.Interpreter(model_path=str(model_path))
        self.model_interpreter.allocate_tensors()

        feature_path = self.directory / "feature_extractor.tflite"
        if not feature_path.exists():
            warnings.warn(
                f"Could not find feature extractor model file ({feature_path}).\n"
                f"Using the tflite runner without a compiled feature extractor is "
                f"an experimental feature and may lead to poor model performance."
            )
            self.feature_interpreter = None
        else:
            self.feature_interpreter = tf.lite.Interpreter(model_path=str(feature_path))
            self.feature_interpreter.allocate_tensors()

        self.reset_state()

    def reset_state(self, batch_size: int = 1) -> None:
        """Reset the internal state of the model to initial conditions."""
        assert self.model_interpreter is not None
        self.model_state = [
            np.zeros([batch_size] + x["shape"].tolist()[1:], dtype=x["dtype"])
            for x in self.model_interpreter.get_input_details()[1:]
        ]

        if self.feature_interpreter is not None:
            self.feature_state = [
                np.zeros(
                    [
                        # initialize dynamic state sizes to 0 to avoid startup artifacts
                        batch_size if i == 0 else (0 if sig == -1 else val)
                        for i, (sig, val) in enumerate(
                            zip(x["shape_signature"], x["shape"])
                        )
                    ],
                    dtype=x["dtype"],
                )
                for x in self.feature_interpreter.get_input_details()[1:]
            ]

    def run(self, inputs: np.ndarray) -> np.ndarray:
        """
        Run the model on the given inputs.

        Parameters
        ----------
        inputs : np.ndarray
            Model input values (should have shape ``(batch_size, input_steps)``).

        Returns
        -------
        outputs : ``np.ndarray``
            Model output values (with shape ``(batch_size, output_d)`` if
            ``self.model_params['return_sequences']=False``
            else ``(batch_size, output_steps, output_d)``).
        """
        if self.feature_interpreter is not None:
            inputs, self.feature_state = self._run_feature_extractor(inputs)

        outputs, self.model_state = self._run_model(inputs)

        return outputs

    def _resize_inputs(
        self, interpreter: "tf.lite.Interpreter", input_vals: List[np.ndarray]
    ) -> None:
        reallocate = False
        for s0, s1 in zip(input_vals, interpreter.get_input_details()):
            if s0.shape != tuple(s1["shape"]):
                # this will attempt to resize any dynamic dimensions (e.g. batch
                # size), and will raise an error if there is a mismatch on
                # static dimensions
                interpreter.resize_tensor_input(s1["index"], s0.shape, strict=True)
                reallocate = True
        if reallocate:
            interpreter.allocate_tensors()

    def _run_feature_extractor(
        self, inputs: np.ndarray
    ) -> Tuple[np.ndarray, List[np.ndarray]]:
        """Run the feature extractor on the given inputs."""
        assert self.feature_interpreter is not None

        self._resize_inputs(self.feature_interpreter, [inputs] + self.feature_state)

        for inp, val in zip(
            self.feature_interpreter.get_input_details(), [inputs] + self.feature_state
        ):
            self.feature_interpreter.set_tensor(inp["index"], val.astype(np.float32))

        self.feature_interpreter.invoke()

        outputs = self.feature_interpreter.get_tensor(
            self.feature_interpreter.get_output_details()[0]["index"]
        )
        feature_state = [
            self.feature_interpreter.get_tensor(out["index"])
            for out in self.feature_interpreter.get_output_details()[1:]
        ]

        return outputs, feature_state

    def _run_model(self, inputs: np.ndarray) -> Tuple[np.ndarray, List[np.ndarray]]:
        """Run the main model on the given inputs."""

        n_unroll = self.model_interpreter.get_input_details()[0]["shape"][1]
        n_steps = inputs.shape[1]
        if n_steps % n_unroll != 0:
            raise ValueError(
                f"Number of timesteps ({n_steps}) cannot be evenly divided by unroll "
                f"size ({n_unroll})"
            )

        self._resize_inputs(
            self.model_interpreter, [inputs[:, :n_unroll]] + self.model_state
        )
        input_details = self.model_interpreter.get_input_details()
        output_details = self.model_interpreter.get_output_details()

        input_tensors = [
            self.model_interpreter.tensor(details["index"]) for details in input_details
        ]
        output_tensors = [
            self.model_interpreter.tensor(details["index"])
            for details in output_details
        ]

        # quantize all the inputs ahead of time (for efficiency)
        inputs = self.quantize(inputs, input_details[0])

        state_input = self.model_state
        if self.model_params["return_sequences"]:
            output_sequences = []

        for step in range(0, n_steps, n_unroll):
            # set input for current block of unrolled steps
            input_tensors[0]()[:] = inputs[:, step : step + n_unroll]
            for inp, fp32_data, details in zip(
                input_tensors[1:], state_input, input_details[1:]
            ):
                inp()[:] = self.quantize(fp32_data, details)

            # run interpreter
            self.model_interpreter.invoke()

            # get state input for next block of steps
            state_input = [
                self.dequantize(out(), details)
                for out, details in zip(output_tensors[1:], output_details[1:])
            ]

            if self.model_params["return_sequences"]:
                # Append all outputs from the past `unroll` timesteps
                output_sequences.append(output_tensors[0]().copy())

        if self.model_params["return_sequences"]:
            # concatenate outputs from each unroll block
            # note: because the model was built with return_sequences=True, this
            # represents the output from all the timesteps
            outputs = np.concatenate(output_sequences, axis=1)
        else:
            # only use the output from last unroll block
            # note: because the model was built with return_sequences=False, the
            # output from the last unroll block is actually the output from the
            # last timestep
            outputs = output_tensors[0]().copy()

        # dequantize outputs (state_input was already dequantized in the loop above)
        outputs = self.dequantize(outputs, output_details[0])

        return outputs, state_input

    @staticmethod
    def quantize(val: np.ndarray, details: Dict[str, Any]) -> np.ndarray:
        """Quantize the given value based on quantization parameters."""
        n_scales = len(details["quantization_parameters"]["scales"])
        if n_scales == 0:
            # val is not quantized
            return np.array(val)

        assert n_scales == 1
        val = np.asarray(val)
        scaling_factor = details["quantization_parameters"]["scales"][0]
        zero_point = details["quantization_parameters"]["zero_points"][0]
        iinfo = np.iinfo(details["dtype"])
        q = np.round((val / scaling_factor) + zero_point)
        return np.clip(q, iinfo.min, iinfo.max).astype(details["dtype"])

    @staticmethod
    def dequantize(val: np.ndarray, details: Dict[str, Any]) -> np.ndarray:
        """Dequantize the given value based on quantization parameters."""
        n_scales = len(details["quantization_parameters"]["scales"])
        if n_scales == 0:
            # val is not quantized
            return np.array(val)

        assert n_scales == 1
        scaling_factor = details["quantization_parameters"]["scales"][0]
        zero_point = details["quantization_parameters"]["zero_points"][0]
        return (val.astype(np.float32) - zero_point) * scaling_factor



# pylint: disable=missing-docstring

from typing import Iterable

import numpy as np
import pytest
import tensorflow as tf

from nengo_edge import ragged


@pytest.mark.parametrize(
    "val",
    (
        [[0, 1, 2], [3, 4]],
        np.array([[0, 1, 2], [3, 4]], dtype=object),
        tf.zeros((2, 3)),
        tf.ragged.constant([[0, 1, 2], [3, 4]]),
    ),
)
def test_to_from_masked(val: Iterable) -> None:
    masked = ragged.to_masked(val)
    assert np.array_equal(
        tf.math.count_nonzero(masked._keras_mask, axis=-1), [len(x) for x in val]
    )

    unmasked = ragged.from_masked(masked)
    assert np.array_equal([len(x) for x in unmasked], [len(x) for x in val])



"""Tools for working with ragged arrays in Numpy/TensorFlow."""

from typing import Iterable, Union

import tensorflow as tf


def to_tf(x: Iterable) -> tf.RaggedTensor:
    """Convert inputs to a TensorFlow RaggedTensor."""
    if isinstance(x, tf.RaggedTensor):
        return x
    elif isinstance(x, tf.Tensor):
        return tf.RaggedTensor.from_tensor(x)
    return tf.ragged.constant(x)


def to_masked(x: Iterable) -> tf.Tensor:
    """Convert inputs to a masked TensorFlow Tensor."""
    ragged_x = to_tf(x)
    mask = tf.sequence_mask(ragged_x.row_lengths())
    masked_x = ragged_x.to_tensor()
    masked_x._keras_mask = mask
    return masked_x


def from_masked(x: tf.Tensor) -> Union[tf.Tensor, tf.RaggedTensor]:
    """Convert from a masked Tensor to a RaggedTensor."""

    if tf.reduce_all(getattr(x, "_keras_mask", True)):
        # Mask is all True (or tensor is not masked), so just return as-is
        return x

    return tf.ragged.boolean_mask(x, x._keras_mask)



"""Command-line interface for nengo-edge."""

import tarfile
from pathlib import Path
from typing import Optional

import click
from rich import progress

from nengo_edge import validate
from nengo_edge.version import version


@click.group()
def cli() -> None:
    """CLI for nengo-edge."""


@cli.command()
@click.argument(
    "dataset_type",
    type=click.Choice(["kws", "asr"], case_sensitive=False),
)
@click.argument(
    "data_directory",
    type=click.Path(exists=True, dir_okay=True),
)
@click.option(
    "--validation-samples",
    default=None,
    type=int,
    help=(
        "Set the number of audio files to be validated before the dataset is packaged."
    ),
)
def package_dataset(
    dataset_type: str,
    data_directory: str,
    validation_samples: Optional[int],
) -> None:
    """
    Validate and package a dataset for a given model type.

    Parameters
    ----------
    dataset_type : str
        One of kws or asr
    data_directory : str
        Directory containing a dataset split into train, validation and test
        subdirectories.
    validation_samples : Optional[int]
        Maximum number of audio files to validate before packaging the dataset.
    """
    # use resolved data_directory to avoid "../" or symlinks
    data_directory_path = Path(data_directory).resolve()
    tar_fname = (
        Path.cwd() / f"{dataset_type}-v{version}-{data_directory_path.name}.tar.gz"
    )
    # throw error if file already exists in cwd
    if tar_fname.exists():
        raise FileExistsError(f"{tar_fname} already exists.")

    # get validated filepaths
    file_paths = validate.validate_dataset(
        dataset_type.lower(), data_directory_path, validation_samples
    )

    with tarfile.open(tar_fname, mode="w:gz") as f:
        for path in progress.track(
            file_paths,
            description=f"Compressing dataset ({len(file_paths)} entries)",
        ):
            # use arcname to set filepath inside the archive to the
            # relative path with respect to data_directory_path
            f.add(str(path), arcname=str(path.relative_to(data_directory_path)))

    print(f"Saved {dataset_type} dataset archive to {tar_fname}")



"""Interface for loading parameter file associated with a compiled model and runner."""

import json
import warnings
from pathlib import Path
from typing import Any, Dict, Tuple

import packaging

from nengo_edge.version import version


def load_params(directory: Path) -> Tuple[Dict[str, Any], ...]:
    """Load parameters from file."""

    param_path = directory / "parameters.json"
    if not param_path.exists():
        raise FileNotFoundError(f"Could not find parameter file ({param_path})")
    with open(param_path, "r", encoding="utf-8") as f:
        params = json.load(f)
        model_params = params["model"]
        preprocessing = params["preprocessing"]
        postprocessing = params["postprocessing"]

    check_params(params)
    return model_params, preprocessing, postprocessing


def check_params(params: Dict[str, Any]) -> None:
    """Perform validation checks on parameters."""

    # Warn user if the nengo-edge version that was used when hardware artifacts were
    # compiled does not match the user's local nengo-edge version
    expected_version = params["version"]["nengo-edge"]
    if packaging.version.parse(expected_version) != packaging.version.parse(version):
        local_version = packaging.version.parse(version)
        warnings.warn(
            f"Downloaded model uses nengo-edge "
            f"{expected_version}, but you're using nengo-edge {local_version}. "
            f"Mismatch may cause errors."
        )



# pylint: disable=missing-docstring

import random
from pathlib import Path
from typing import Any, Dict, List, Tuple

import numpy as np
import pytest
import soundfile

from nengo_edge import validate


def save_audio_file(
    filepath: Path,
    audio_type: str,
    sample_rate: int,
    duration: float,
) -> None:
    # note: creating 2 channel data, second channel will be ignored
    sample_data = np.random.uniform(
        -32000, 32000, size=(int(sample_rate * duration), 2)
    ).astype("int16")
    soundfile.write(filepath, sample_data, sample_rate, format=audio_type)


def save_transcript_file(
    filepath: Path, fileIDs: List[str], sentences: List[str]
) -> None:
    with open(filepath, "w", encoding="utf-8") as f:
        for fileID, sentence in zip(fileIDs, sentences):
            f.write(f"{fileID} {sentence}\n")


def mock_asr_dataset(
    tmp_path: Path,
    audio_type: str = "wav",
    sample_rate: int = 16000,
    duration: float = 0.5,  # in seconds
    how_many: int = 10,
) -> None:
    rng = random.Random(0)
    for split in ["train", "validation", "test"]:
        split_dir = tmp_path / split
        split_dir.mkdir()
        fileIDs = []
        sentences = []
        for i in range(how_many):
            fileIDs.append(f"some_audio_{i}.{audio_type}")
            sentences.append(f"{''.join(rng.choices('abc ' , k=100))}")
            save_audio_file(
                split_dir / f"some_audio_{i}.{audio_type}",
                audio_type=audio_type,
                sample_rate=sample_rate,
                duration=duration,
            )
        save_transcript_file(split_dir / "transcription.txt", fileIDs, sentences)


def mock_kws_dataset(
    tmp_path: Path,
    audio_type: str = "wav",
    sample_rate: int = 16000,
    duration: float = 0.5,  # in seconds
    keywords: Tuple[str, ...] = ("on", "off"),
    how_many: int = 10,
) -> None:
    for split in ["train", "validation", "test"]:
        split_dir = tmp_path / split
        for word in keywords:
            keyword_dir = split_dir / word
            keyword_dir.mkdir(parents=True)
            for i in range(how_many):
                save_audio_file(
                    keyword_dir / f"some_audio_{i}.{audio_type}",
                    audio_type=audio_type,
                    sample_rate=sample_rate,
                    duration=duration,
                )

    background_dir = tmp_path / "background_noise"
    background_dir.mkdir()
    for i in range(2):
        save_audio_file(
            background_dir / f"some_audio_{i}.{audio_type}",
            audio_type=audio_type,
            sample_rate=sample_rate,
            duration=duration,
        )


@pytest.mark.parametrize(
    "split",
    ("train", "validation", "test", "unknown"),
)
def test_infer_split(split: str) -> None:
    path = Path(f"{split}/path/to/file.txt")
    if split == "unknown":
        with pytest.raises(ValueError, match=f"{path} should be"):
            validate.infer_split(path)
    else:
        assert split == validate.infer_split(path)


@pytest.mark.parametrize("audio_format", ("wav", "flac", "mp3"))
def test_infer_audio_properties(audio_format: str, tmp_path: Path) -> None:
    sample_rate = 16000
    duration = 0.5
    save_audio_file(
        tmp_path / f"audio.{audio_format}",
        audio_type=audio_format,
        sample_rate=sample_rate,
        duration=duration,
    )

    with open(tmp_path / f"audio.{audio_format}", "rb") as f:
        audio_bytes = f.read()
        audio_properties = validate.infer_audio_properties(audio_bytes)

    assert audio_properties["sample_rate"] == sample_rate

    # validate.infer_audio_properties returns clip duration in ms
    assert audio_properties["clip_duration"] == duration * 1e3


def test_validate_audio(tmp_path: Path) -> None:
    sample_rate = 16000
    duration = 0.5

    # test clip duration warning
    data_dir = tmp_path / "kws0"
    data_dir.mkdir()
    mock_kws_dataset(data_dir, sample_rate=sample_rate, duration=duration)
    save_audio_file(
        data_dir / "audio.wav",
        audio_type="wav",
        sample_rate=sample_rate,
        duration=duration + 10,
    )

    audio_files = sorted(list(data_dir.rglob("*.wav")))
    with pytest.warns(match="Clip durations"):
        validate.validate_audio(audio_files)

    # test sample rate error
    data_dir = tmp_path / "kws1"
    data_dir.mkdir()
    mock_kws_dataset(data_dir, sample_rate=sample_rate, duration=duration)
    save_audio_file(
        data_dir / "audio.wav",
        audio_type="wav",
        sample_rate=sample_rate + 1,
        duration=duration,
    )

    audio_files = list(data_dir.rglob("*.wav"))
    with pytest.raises(ValueError, match="Found sample rates"):
        validate.validate_audio(audio_files)


def test_audio_map_fn(tmp_path: Path) -> None:
    save_audio_file(
        tmp_path / "audio.wav", audio_type="wav", sample_rate=16000, duration=1
    )

    props = validate.audio_map_fn(tmp_path / "audio.wav")
    assert props["sample_rate"] == 16000
    assert props["clip_duration"] == 1000


def test_validate_transcriptions(tmp_path: Path) -> None:
    data_dir = tmp_path / "asr0"
    data_dir.mkdir()
    mock_asr_dataset(data_dir)

    with open(data_dir / "train" / "transcription.txt", "a+", encoding="utf-8") as f:
        f.write("dummy_id.audio dummy sentence\n")

    audio_files = list(data_dir.rglob("*.wav"))

    with open(data_dir / "train" / "transcription.txt", "r", encoding="utf-8") as f:
        transcription_lines = [
            f"{data_dir / 'train'}/{line}" for line in f.read().splitlines()
        ]

    with pytest.raises(
        FileNotFoundError,
        match=f"reference to {data_dir / 'train' / 'dummy_id.audio'}",
    ):
        validate.validate_transcriptions(transcription_lines, audio_files)


def test_validate_kws_formatting(tmp_path: Path) -> None:
    # test incompatible audio format
    with pytest.raises(FileNotFoundError, match="Found no compatible audio"):
        validate.validate_kws_formatting([])

    # test missing split
    fpaths = [
        tmp_path / f"{split}" / "keyword" / "some_audio.wav"
        for split in ("train", "validation")
    ]
    with pytest.raises(FileNotFoundError, match="Did not find train/"):
        validate.validate_kws_formatting(fpaths)

    # test warning when no background_noise dir present
    fpaths = [
        tmp_path / f"{split}" / "keyword" / "some_audio.wav"
        for split in ("train", "validation", "test")
    ]
    with pytest.warns(match="No background_noise"):
        validate.validate_kws_formatting(fpaths)

    # test warning when extra top level directory is found with
    # _dirname_ format
    fpaths = [
        tmp_path / f"{split}" / "keyword" / "some_audio.wav"
        for split in ("train", "validation", "test", "_invalid_")
    ]
    with pytest.warns(match="Found directory _invalid_"):
        validate.validate_kws_formatting(fpaths)

    # test invalid/no keywords
    fpaths = [
        tmp_path / f"{split}" / "_invalid_" / "some_audio.wav"
        for split in ("train", "validation", "test")
    ]
    with pytest.raises(FileNotFoundError, match="keyword directories"):
        validate.validate_kws_formatting(fpaths)

    # test keyword returns
    fpaths = [
        tmp_path / f"{split}" / f"keyword_{i}" / "some_audio.wav"
        for i, split in enumerate(("train", "validation", "test"))
    ]
    props = validate.validate_kws_formatting(fpaths)
    assert props["allKeywords"] == [f"keyword_{i}" for i in range(3)]


def test_validate_asr_formatting(tmp_path: Path) -> None:
    # test incompatible audio format
    with pytest.raises(FileNotFoundError, match="Found no compatible audio"):
        validate.validate_asr_formatting([], [tmp_path / "some.txt"])

    with pytest.raises(FileNotFoundError, match="Found no compatible transcription"):
        validate.validate_asr_formatting([tmp_path / "some_audio.wav"], [])

    audio = [
        tmp_path / f"{split}" / "some_audio.wav" for split in ("train", "validation")
    ]
    txt = [f.with_suffix(".txt") for f in audio]
    # test missing split
    with pytest.raises(FileNotFoundError, match="Did not find train/"):
        validate.validate_asr_formatting(audio, txt)


def test_validate_dataset_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    # test empty directory error
    with pytest.raises(FileNotFoundError, match="Found empty"):
        validate.validate_dataset("kws", tmp_path, validation_samples=None)

    # test really large dataset warning
    def mock_kws_dataset_validation(audio_paths: List[Path]) -> Dict[str, List[str]]:
        """fast kws validation."""
        return {"allKeywords": []}

    def mock_audio_validation(audio_paths: List[Path], dataset_type: str) -> None:
        """fast audio validation."""

    def rglob(*_: Any) -> List[Path]:
        return [Path(f"audio_{i}.wav") for i in range(int(1e5 + 1))]

    def is_file(*_: Any) -> bool:
        return True

    monkeypatch.setattr(
        validate, "validate_kws_formatting", mock_kws_dataset_validation
    )
    monkeypatch.setattr(validate, "validate_audio", mock_audio_validation)
    monkeypatch.setattr(Path, "rglob", rglob)
    monkeypatch.setattr(Path, "is_file", is_file)

    with pytest.warns(match=f"Validating {int(1e5) + 1} audio files"):
        validate.validate_dataset("kws", tmp_path, validation_samples=None)


@pytest.mark.parametrize("dataset_type", ("kws", "asr"))
def test_validate_dataset(
    dataset_type: str, tmp_path: Path, capfd: pytest.CaptureFixture
) -> None:
    data_dir = tmp_path / dataset_type
    data_dir.mkdir()

    if dataset_type == "kws":
        mock_kws_dataset(data_dir)
    else:
        mock_asr_dataset(data_dir)

    validated_fpaths = validate.validate_dataset(
        dataset_type, data_dir, validation_samples=10
    )

    out = capfd.readouterr().out
    assert "Successfully validated 10 audio files" in out
    if dataset_type == "asr":
        assert "Successfully validated 30 transcription entries" in out
    assert validated_fpaths == [f for f in data_dir.rglob("*") if f.is_file()]



# pylint: disable=missing-docstring

from pathlib import Path

import pytest
from click.testing import CliRunner

from nengo_edge import cli
from nengo_edge.tests.test_validate import mock_asr_dataset, mock_kws_dataset
from nengo_edge.version import version


@pytest.mark.parametrize("dataset_type", ("kws", "asr"))
def test_package_dataset(
    dataset_type: str,
    tmp_path: Path,
    cli_runner: CliRunner,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    data_directory = tmp_path / dataset_type
    data_directory.mkdir()

    if dataset_type == "kws":
        mock_kws_dataset(data_directory)

    elif dataset_type == "asr":
        mock_asr_dataset(data_directory)

    monkeypatch.chdir(tmp_path)
    result = cli_runner.invoke(
        cli.cli,
        f"package-dataset {dataset_type} {data_directory}".split(),
    )
    assert result.exit_code == 0, result.exception

    output_tar = Path.cwd() / f"{dataset_type}-v{version}-{data_directory.name}.tar.gz"
    assert output_tar.exists()

    with pytest.raises(FileExistsError, match="already exists"):
        result = cli_runner.invoke(
            cli.cli,
            f"package-dataset {dataset_type} {data_directory}".split(),
            catch_exceptions=False,
        )



"""Interface for running an exported NengoEdge model in SavedModel format."""

import warnings
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np

from nengo_edge import config

try:
    import tensorflow as tf
    from tensorflow_text import SentencepieceTokenizer

    from nengo_edge import ragged  # pylint: disable=ungrouped-imports

except ImportError:  # pragma: no cover
    warnings.warn("TensorFlow is not installed; cannot use SavedModelRunner.")


class SavedModelRunner:
    """Run a model exported in TensorFlow's SavedModel format."""

    def __init__(self, directory: Union[str, Path]):
        self.directory = Path(directory)

        self.model = tf.keras.saving.load_model(directory, compile=False)

        self.model_params, self.preprocessing, self.postprocessing = config.load_params(
            self.directory
        )

        if self.model_params["type"] == "asr":
            self.tokenizer = self.get_tokenizer(self.postprocessing)
            if self.tokenizer is None:
                warnings.warn(
                    "No tokenizer model found, cannot decode ASR outputs. "
                    "Consider re-downloading ASR run artifacts."
                )
        elif self.model_params["type"] == "nlp":
            self.tokenizer = self.get_tokenizer(self.preprocessing)
            if self.tokenizer is None:
                raise ValueError(
                    "No tokenizer model found and is required to process string inputs."
                    "Please re-download the NLP artifacts."
                )

    def get_tokenizer(
        self, params: Dict[str, str]
    ) -> Optional["SentencepieceTokenizer"]:
        """
        Load a ``tensorflow_text.SentencepieceTokenizer``.

        Returns None if no ``tokenizer_file`` exists in the load directory.

        Parameters
        ----------
        params: Dict[str, str]
            Configuration dictionary containing the name of the
            ``tokenizer_file`` found in the save directory.

        Returns
        -------
        tokenizer: ``tensorflow_text.SentencepieceTokenizer``
            A sentencepiece tokenizer with a trained vocabulary.
        """
        if (
            "tokenizer_file" in params
            and (self.directory / params["tokenizer_file"]).exists()
        ):
            return SentencepieceTokenizer(
                (self.directory / params["tokenizer_file"]).read_bytes()
            )
        return None

    def _run_model(self, inputs: np.ndarray) -> np.ndarray:
        """
        Run the main model logic on the given inputs.

        Parameters
        ----------
        inputs : np.ndarray
            Model input values (should have shape ``(batch_size, input_steps)``).

        Returns
        -------
        outputs : NDarray
            Model output values in a ragged tensor.
        """

        ragged_inputs = ragged.to_tf(inputs)
        ragged_inputs = tf.cast(ragged_inputs, "float32")
        masked_inputs = ragged.to_masked(ragged_inputs)

        model_inputs = tf.nest.flatten(masked_inputs)

        output = tf.nest.flatten(self.model(model_inputs))[0]

        return ragged.from_masked(output).numpy()

    def run(self, inputs: Union[np.ndarray, List[str]]) -> np.ndarray:
        """
        Run the model on the given inputs.

        This function applies model specific actions to inputs or outputs.
        For example, ASR model outputs are automatically detokenized to construct
        a string prediction. For NLP models, the list of strings is
        automatically processed into integer tokens from a saved
        vocabulary.

        Parameters
        ----------
        inputs : Union[np.ndarray, List[str]]
            For audio based models, such as KWS and ASR, model input values
            should be stored in a numpy array with a shape of
            ``(batch_size, input_steps)``.
            For text based NLP models, input strings should be stored in a list with
            a length equal to ``batch_size``.

        Returns
        -------
        outputs : np.ndarray
            Model output values with shape ``(batch_size, output_steps, output_d)``
            for KWS, ``(batch_size)`` for ASR and ``(batch_size, output_d)`` for NLP.
        """

        if isinstance(inputs[0], str):
            # Process string inputs
            assert self.model_params["type"] == "nlp"
            assert self.tokenizer is not None
            inputs = self.tokenizer.tokenize([s.lower() for s in inputs]).numpy()

        assert isinstance(inputs, np.ndarray)
        if self.model_params["type"] == "kws" and inputs.dtype == object:
            raise NotImplementedError("KWS models do not support ragged inputs")

        outputs = self._run_model(inputs)

        # Detokenize asr outputs
        if self.model_params["type"] == "asr" and self.tokenizer is not None:
            # Apply CTC decoding
            outputs = tf.keras.backend.ctc_decode(
                tf.nn.softmax(ragged.to_masked(outputs)),
                input_length=[len(x) for x in outputs],
                greedy=True,
            )[0][0]
            outputs = tf.cast(outputs, "int32")

            # Filter out blank/unknown tokens
            outputs = tf.ragged.boolean_mask(outputs, outputs > 0)

            # Convert tokens back to text
            outputs = np.asarray(
                [s.decode("utf-8") for s in self.tokenizer.detokenize(outputs).numpy()]
            )

        return outputs



# pylint: disable=missing-docstring

import warnings
from pathlib import Path

import pytest

from nengo_edge import config, version


def test_config_errors(tmp_path: Path) -> None:
    with pytest.raises(FileNotFoundError, match="Could not find parameter file"):
        config.load_params(tmp_path)


def test_hw_artifact_compatibility() -> None:
    # assert warning case
    expected_hw_version = "0.0.0"
    with pytest.warns(UserWarning, match="Mismatch may cause errors."):
        config.check_params({"version": {"nengo-edge": expected_hw_version}})

    # assert no warning, passing case
    expected_hw_version = version.version
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        config.check_params({"version": {"nengo-edge": expected_hw_version}})



import json
from typing import Any

import pytest
from click.testing import CliRunner

from nengo_edge.version import version


@pytest.fixture
def param_dir(tmp_path):
    params = {
        "preprocessing": {
            "window_size_ms": 40,
            "window_stride_ms": 20,
            "mel_num_bins": 40,
            "dct_num_features": 20,
            "sample_rate": 16000,
            "mel_lower_edge_hertz": 20,
            "mel_upper_edge_hertz": 7000,
            "log_epsilon": 1e-12,
        },
        "model": {
            "input_shape": [None, 20],
            "output_shape": [None, 10],
            "state_shapes": [10, 10],
            "return_sequences": True,
            "n_unroll": 1,
            "type": "kws",
        },
        "postprocessing": {
            "vocab_size": 20,
            "tokenizer_file": "sentencepiece_20.model",
        },
        "version": {"nengo-edge": version},
    }

    with open(tmp_path / "parameters.json", "w", encoding="utf-8") as f:
        json.dump(params, f)

    return tmp_path


@pytest.fixture
def cli_runner() -> CliRunner:
    """
    A version of ``click.testing.CliRunner`` that echoes output to stdout.

    This is useful as it allows pytest to capture the output like normal (rather
    than the CliRunner eating it all).
    """

    class EchoCliRunner(CliRunner):
        def invoke(self, *args: Any, **kwargs: Any) -> Any:
            result = super().invoke(*args, **kwargs)
            print(result.stdout)
            return result

    return EchoCliRunner()


def pytest_sessionstart(session: pytest.Session) -> None:
    """Hook that runs on session start."""

    try:
        import tensorflow as tf  # pylint: disable=unused-import
    except ImportError:  # pragma: no cover
        pytest.exit("Tensorflow is not installed; cannot run tests.")



"""
Interface for running an exported NengoEdge model on network accessible devices.

Nengo-edge supports running on the Coral dev board via this runner.
"""

import subprocess
import tempfile
from pathlib import Path
from typing import List, Optional, Union

import numpy as np

from nengo_edge import config
from nengo_edge.device_modules import coral_device, np_mfcc


class NetworkRunner:
    """
    Run an exported component on a remote network device.

    Parameters
    ----------
    directory : Union[str, Path]
        Path to the directory containing the NengoEdge export artifacts.
    username : str
        Username on the remote device
    hostname : str
        Hostname of the remote device
    local : bool
        If True, run locally rather than over the network (for debugging purposes).
    """

    def __init__(
        self,
        directory: Union[str, Path],
        username: str,
        hostname: str,
        local: bool = False,
    ):
        self.directory = Path(directory)
        self.model_params, self.preprocessing, self.postprocessing = config.load_params(
            self.directory
        )
        self.username = username
        self.hostname = hostname
        self.address = f"{self.username}@{self.hostname}"
        self.local = local

        try:
            self._ssh("echo ok")
        except Exception as e:
            raise RuntimeError(
                f"Cannot connect to address {self.address}: {e!r}"
            ) from e

        if local:
            self._tmp_dir = (
                tempfile.TemporaryDirectory(  # pylint: disable=consider-using-with
                    dir=directory
                )
            )
            self.remote_dir = Path(self._tmp_dir.name)
        else:  # pragma: no cover (needs device)
            self.remote_dir = Path(f"/tmp/nengo-edge-{type(self).__name__.lower()}")
            self._ssh(f"rm -rf {self.remote_dir}")
            self._ssh(f"mkdir -p {self.remote_dir}")

    def _scp(self, files_to_copy: List[Path]) -> None:
        """One liner to send specified files to remote device location."""
        cmd = (
            (["cp", "-r"] if self.local else ["scp"])
            + [str(p) for p in files_to_copy]
            + [
                (
                    str(self.remote_dir)
                    if self.local
                    else f"{self.address}:{self.remote_dir}"
                )
            ]
        )

        subprocess.run(cmd, check=True)

    def _ssh(self, cmd: str, std_in: Optional[bytes] = None) -> bytes:
        """Run a command over ssh."""
        return subprocess.run(
            cmd if self.local else ["ssh", self.address, cmd],
            input=std_in,
            check=True,
            stdout=subprocess.PIPE,
            shell=self.local,
        ).stdout


class CoralRunner(NetworkRunner):
    """
    Run a model exported from NengoEdge on the Coral board.

    See `NetworkRunner` for parameter descriptions.
    """

    def __init__(
        self,
        directory: Union[str, Path],
        username: str,
        hostname: str,
        local: bool = False,
    ):
        super().__init__(
            directory=directory, username=username, hostname=hostname, local=local
        )

        self.device_runner = Path(coral_device.__file__)
        self.return_sequences = self.model_params["return_sequences"]

        # copy files to remote
        self._scp(
            [
                self.device_runner,
                self.directory / "model_edgetpu.tflite",
                self.directory / "parameters.json",
                Path(np_mfcc.__file__),
            ]
        )

    def run(
        self,
        inputs: np.ndarray,
    ) -> np.ndarray:  # pragma: no cover (needs device)
        """
        Run model inference on a batch of inputs.

        Parameters
        ----------
        inputs : np.ndarray
            Model input values (must have shape ``(batch_size, input_steps)``)

        Returns
        -------
        outputs : ``np.ndarray``
            Model output values (with shape ``(batch_size, output_d)`` if
            ``self.model_params['return_sequences']=False``
            else ``(batch_size, output_steps, output_d)``).
        """

        # Save inputs to file
        filepath = self.directory / "inputs.npz"
        np.savez_compressed(filepath, inputs=inputs)

        # Copy to device
        self._scp([filepath])

        # Run model on device
        self._ssh(
            f"python3 {self.remote_dir / self.device_runner.name} "
            f"--directory {self.remote_dir} "
            f"{'--return-sequences' if self.return_sequences else ''}"
        )

        # Copy outputs back to host
        subprocess.run(
            f"scp {self.address}:{self.remote_dir / 'outputs.npy'} "
            f"{self.directory}".split(),
            check=True,
        )

        outputs = np.load(self.directory / "outputs.npy")
        return outputs


class NetworkTokenizer(NetworkRunner):
    """
    Class to access a SentencePiece command line interface installed on a network
    accessible device.

    Can be used to tokenize and detokenize values for asr and nlp inference.
    """

    def __init__(
        self,
        directory: Union[str, Path],
        username: str,
        hostname: str,
        local: bool = False,
    ):
        super().__init__(
            directory=directory, username=username, hostname=hostname, local=local
        )

        self.tokenizer_file = None
        for params in [self.preprocessing, self.postprocessing]:
            if "tokenizer_file" in params:
                self.tokenizer_file = params["tokenizer_file"]

        if self.tokenizer_file is None:
            raise TypeError("Exported config does not contain any tokenizers")

        # Copy files to remote
        self._scp([self.directory / self.tokenizer_file])

    def tokenize(self, input_text: str) -> List[int]:
        """
        Map strings to their corresponding integer tokens.

        This function utilizes an ssh command to access the SentencePiece command
        line interface on the configured network device.

        Parameters
        ----------
        input_text: str
            Input string to be tokenized.

        Returns
        -------
        token_ids: List[int]
            A list of integers of length ``(n_tokens)``.
        """
        assert self.tokenizer_file is not None
        output = self._ssh(
            f"spm_encode"
            f" --model={self.remote_dir / self.tokenizer_file}"
            f" --output_format=id",
            std_in=input_text.encode("utf-8"),
        ).decode("utf-8")
        token_string = output.rstrip()
        token_ids = [int(token) for token in token_string.split()]
        return token_ids

    def detokenize(self, inputs: np.ndarray) -> str:
        """
        Map integer tokens to their corresponding string token.

        This function utilizes an ssh command to access the SentencePiece command
        line interface on the configured network device.

        Parameters
        ----------
        inputs: np.ndarray
            Input array containing integer values. Array should be generated
            from a top-1 decoding strategy (e.g. greedy decoding)
            on asr model outputs and have a size of ``(batch_size, output_steps)``.

        Returns
        -------
        decoded_text: str
            A string generated from the decoded input integers.
        """

        token_string = " ".join([str(token) for token in inputs[inputs != -1]])

        assert self.tokenizer_file is not None
        output = self._ssh(
            f"spm_decode"
            f" --model={self.remote_dir / self.tokenizer_file}"
            f" --input_format=id",
            std_in=token_string.encode("utf-8"),
        ).decode("utf-8")
        decoded_text = output.rstrip()
        return decoded_text

    def run(self, inputs: Union[np.ndarray, List[str]]) -> np.ndarray:
        """Run the main tokenizer logic on the given inputs."""

        if isinstance(inputs[0], str):
            outputs = np.asarray([self.tokenize(text) for text in inputs], dtype=object)
        else:
            assert isinstance(inputs, np.ndarray)

            if inputs.dtype not in ["int32", "int64"]:
                raise ValueError(f"{inputs.dtype=} must be one of int32/int64.")

            if inputs.ndim != 2:
                raise ValueError(
                    f"inputs must have exactly 2 dimensions, found {inputs.ndim}."
                )

            outputs = np.asarray(
                [self.detokenize(tokens) for tokens in inputs], dtype=np.str_
            )
        return outputs



# pylint: disable=missing-docstring

import json
from pathlib import Path
from string import ascii_lowercase
from typing import Any, Dict, Literal, Tuple

import numpy as np
import pytest
import tensorflow as tf
from nengo_edge_hw import gpu
from nengo_edge_models.asr.metrics import decode_predictions
from nengo_edge_models.asr.models import asr_tiny
from nengo_edge_models.kws.models import kws_small
from nengo_edge_models.layers import Tokenizer
from nengo_edge_models.models import Tokenizer as TokenizerDesc
from nengo_edge_models.nlp.models import LMUformerEncoderNLP, nlp_sim_small

from nengo_edge import ragged, version
from nengo_edge.saved_model_runner import SavedModelRunner


def new_tokenizer(tmp_path: Path) -> Tuple[Tokenizer, Path]:
    tokenizer = Tokenizer(
        vocab_size=256,
        corpus=Path(__file__).read_text(encoding="utf-8").splitlines(),
    )
    (tmp_path / "tokenizer").mkdir(parents=True, exist_ok=True)
    tokenizer_path = tokenizer.save(tmp_path / "tokenizer")
    return tokenizer, tokenizer_path


@pytest.mark.parametrize("mode", ("model-only", "feature-only", "full"))
def test_runner(
    mode: Literal["model-only", "feature-only", "full"],
    rng: np.random.RandomState,
    seed: int,
    tmp_path: Path,
) -> None:
    tf.keras.utils.set_random_seed(seed)

    pipeline = kws_small()
    if mode == "feature-only":
        pipeline.model = []
    elif mode == "model-only":
        pipeline.pre = []

    interface = gpu.host.Interface(pipeline, build_dir=tmp_path)

    inputs = rng.uniform(
        -1, 1, size=(32,) + ((49, 20) if mode == "model-only" else (16000,))
    ).astype("float32")

    output0 = interface.run(inputs)

    interface.export_model(tmp_path)
    runner = SavedModelRunner(tmp_path)

    output1 = runner.run(inputs)

    assert np.allclose(output0, output1), np.max(np.abs(output0 - output1))


@pytest.mark.parametrize("mode", ("model-only", "feature-only", "full"))
@pytest.mark.parametrize("model_type", ("asr", "nlp"))
def test_runner_ragged(
    mode: str, model_type: str, rng: np.random.RandomState, seed: int, tmp_path: Path
) -> None:
    tf.keras.utils.set_random_seed(seed)

    pipeline = asr_tiny() if model_type == "asr" else nlp_sim_small()
    _, tokenizer_path = new_tokenizer(tmp_path)

    if model_type == "asr":
        assert isinstance(pipeline.post[-1], TokenizerDesc)
        pipeline.post[-1].tokenizer_file = tokenizer_path
    else:
        assert isinstance(pipeline.pre[0], TokenizerDesc)
        assert isinstance(pipeline.model[1], LMUformerEncoderNLP)
        pipeline.pre[0].tokenizer_file = tokenizer_path
        pipeline.pre[0].vocab_size = 256
        pipeline.model[1].vocab_size = 256

    if mode == "feature-only":
        pipeline.model = []
        pipeline.post = []

    elif mode == "model-only":
        if model_type == "asr":
            pipeline.pre = []
        pipeline.post = []

    interface = gpu.host.Interface(pipeline, build_dir=tmp_path, return_sequences=True)
    interface.export_model(tmp_path)
    runner = SavedModelRunner(tmp_path)
    if model_type == "asr":
        inputs = rng.uniform(
            -1, 1, size=(32,) + ((49, 80) if mode == "model-only" else (16000,))
        ).astype("float32")

        inputs = np.array(
            [
                inputs[0, : int(inputs.shape[1] * 0.5)],
                inputs[1, : int(inputs.shape[1] * 0.8)],
            ],
            dtype=object,
        )
        ragged_in0 = inputs[0:1]
        ragged_in1 = inputs[1:2]
    else:
        inputs = np.asarray(
            [
                " ".join(
                    "".join(rng.choice([*ascii_lowercase], size=3))
                    for _ in range(10 + i * 10)
                )
                for i in range(2)
            ]
        )

        ragged_in0 = inputs[[0]]
        ragged_in1 = inputs[[1]]

    ragged_out = runner.run(inputs)
    ragged_out0 = runner.run(ragged_in0)
    ragged_out1 = runner.run(ragged_in1)

    if mode == "full" and model_type == "asr":
        # test string output case
        assert len(ragged_out[0]) == len(ragged_out0[0])
        assert len(ragged_out[1]) == len(ragged_out1[0])
        assert ragged_out[0] == ragged_out0[0]
        assert ragged_out[1] == ragged_out1[0]
    else:
        # test numerical output case
        assert (1,) + ragged_out[0].shape == ragged_out0.shape
        assert (1,) + ragged_out[1].shape == ragged_out1.shape

        assert np.allclose(ragged_out[0][None, ...], ragged_out0, atol=4e-6), np.max(
            abs(ragged_out[0] - ragged_out0)
        )
        assert np.allclose(ragged_out[1][None, ...], ragged_out1, atol=4e-6), np.max(
            abs(ragged_out[1] - ragged_out1)
        )


def test_asr_detokenization(
    rng: np.random.RandomState,
    seed: int,
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    tf.keras.utils.set_random_seed(seed)

    pipeline = asr_tiny()
    tokenizer, tokenizer_path = new_tokenizer(tmp_path)
    assert isinstance(pipeline.post[-1], TokenizerDesc)
    pipeline.post[-1].tokenizer_file = tokenizer_path

    interface = gpu.host.Interface(pipeline, build_dir=tmp_path, return_sequences=True)
    interface.export_model(tmp_path)

    monkeypatch.setattr(SavedModelRunner, "_run_model", lambda s, x: x)
    runner = SavedModelRunner(tmp_path)

    inputs = rng.uniform(0, 1, (32, 100, 257))
    # ensure some blank tokens are in the sequences
    # as well as repeats
    for i in range(0, 32):
        # repeats
        inputs[i, i : i + 2] = i
        # blank at the end of sequence
        inputs[i, -1] = 256

    outputs = runner.run(inputs)
    gt = decode_predictions(ragged.to_masked(inputs), tokenizer)
    assert np.all(outputs == gt)


def test_nlp_tokenization(
    rng: np.random.RandomState,
    seed: int,
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    tf.keras.utils.set_random_seed(seed)

    pipeline = nlp_sim_small()
    tokenizer, tokenizer_path = new_tokenizer(tmp_path)

    assert isinstance(pipeline.pre[0], TokenizerDesc)
    assert isinstance(pipeline.model[1], LMUformerEncoderNLP)
    pipeline.pre[0].tokenizer_file = tokenizer_path
    pipeline.pre[0].vocab_size = 256
    pipeline.model[1].vocab_size = 256

    interface = gpu.host.Interface(pipeline, build_dir=tmp_path)
    interface.export_model(tmp_path)

    monkeypatch.setattr(SavedModelRunner, "_run_model", lambda s, x: x)
    runner = SavedModelRunner(tmp_path)

    text_batch = [
        " ".join("".join(rng.choice([*ascii_lowercase], size=3)) for _ in range(10))
        for _ in range(32)
    ]

    outputs = runner.run(text_batch)
    gt = tokenizer.tokenize(text_batch).numpy()  # type: ignore
    assert len(outputs) == len(gt)
    for x, y in zip(outputs, gt):
        np.testing.assert_allclose(x, y)


def test_runner_error_warnings(tmp_path: Path) -> None:
    # test warning for asr models when no tokenizer file is present
    params: Dict[str, Dict[str, Any]] = {
        "preprocessing": {"tokenizer_file": "does not exist"},
        "postprocessing": {"tokenizer_file": "does not exist"},
        "version": {"nengo-edge": version.version},
        "model": {"type": "asr", "return_sequences": True},
    }

    interface = gpu.host.Interface(
        kws_small(), build_dir=tmp_path, return_sequences=True
    )
    interface.export_model(tmp_path)
    # overwrite exported parameters
    with open(tmp_path / "parameters.json", "w", encoding="utf-8") as f:
        json.dump(params, f)

    with pytest.warns(UserWarning, match="cannot decode ASR outputs"):
        SavedModelRunner(tmp_path)

    # test error for nlp models when no tokenizer file is present
    params["model"]["type"] = "nlp"
    with open(tmp_path / "parameters.json", "w", encoding="utf-8") as f:
        json.dump(params, f)

    with pytest.raises(ValueError, match="required to process string inputs"):
        SavedModelRunner(tmp_path)



# Automatically generated by nengo-bones, do not edit this file directly

import io
import pathlib
import runpy

try:
    from setuptools import find_packages, setup
except ImportError:
    raise ImportError(
        "'setuptools' is required but not installed. To install it, "
        "follow the instructions at "
        "https://pip.pypa.io/en/stable/installing/#installing-with-get-pip-py"
    )


def read(*filenames, **kwargs):
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


root = pathlib.Path(__file__).parent
version = runpy.run_path(str(root / "nengo_edge" / "version.py"))["version"]

install_req = [
    "click>=8.1.3",
    "packaging>=20.9",
    "pyserial>=3.5",
    "rich>=13.3.1",
    "soundfile>=0.12.1",
    "numpy>=1.23.0",
]
docs_req = [
    "jupyter>=1.0.0",
    "nbsphinx>=0.8.11",
    "nengo-sphinx-theme>=20.9",
    "numpydoc>=1.4.0",
    "sounddevice>=0.4.5",
    "sphinx-click>=5.0.0",
    "sphinx-tabs>=3.2.0",
]
optional_req = [
    "tensorflow>=2.10.0",
    "tensorflow-text>=2.13.0",
]
tests_req = [
    "mypy>=0.901",
    "pytest>=7.1.1",
    "pytest-rng>=1.0.0",
    "types-click>=7.1.0",
]

setup(
    name="nengo-edge",
    version=version,
    author="Applied Brain Research",
    author_email="edge-info@appliedbrainresearch.com",
    packages=find_packages(),
    url="https://github.com/nengo/nengo-edge",
    include_package_data=True,
    license="MIT license",
    description="Tools for working with NengoEdge",
    long_description=read("README.rst", "CHANGES.rst"),
    zip_safe=False,
    install_requires=install_req,
    extras_require={
        "all": docs_req + optional_req + tests_req,
        "docs": docs_req,
        "optional": optional_req,
        "tests": tests_req,
    },
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Nengo",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    entry_points={
        "console_scripts": [
            "nengo-edge=nengo_edge.cli:cli",
        ],
    },
)



"""Various helper functions and tools for validating NengoEdge datasets."""

import io
import multiprocessing
import os
import random
import warnings
from pathlib import Path
from typing import Dict, List, Optional, Union

import soundfile
from rich import progress


def infer_split(relative_path: Path) -> str:
    """
    Infer the training split from a Path object.

    The training split is inferred from the base part
    of a file path, where the filepath should be a relative path,
    with respect to its parent data directory. If the inferred split
    is not one of train, validation or test, an error is thrown.

    Parameters
    ----------
    relative_path : Path
        File path, relative to its data directory.

    Returns
    -------
    split : str
        Returns one of train, validation or test.
    """
    for split in ["train", "validation", "test"]:
        if relative_path.parts[0].startswith(split):
            return split

    # raise error if function does not return
    raise ValueError(f"{relative_path} should be in one of train, validation or test")


def infer_audio_properties(audio_bytes: bytes) -> Dict[str, Union[int, float]]:
    """
    Infer clip duration and sample rate from audio bytes.

    Parameters
    ----------
    audio_bytes : bytes
        Bytes like object containing audio to be read

    Returns:
    --------
    props: Dict[str, Union[int, float]]
        Dictionary containing the audio `sample_rate` as an int and
        audio `clip_duration` as a float.
    """
    data, rate = soundfile.read(io.BytesIO(audio_bytes))
    duration = len(data) / rate * 1e3  # in ms
    props = {"clip_duration": duration, "sample_rate": int(rate)}
    return props


def audio_map_fn(file_path: Path) -> Dict[str, Union[int, float]]:
    """Map function for multiprocessing audio files."""
    return infer_audio_properties(file_path.read_bytes())


def validate_audio(
    audio_paths: List[Path], dataset_type: Optional[str] = "kws"
) -> None:
    """
    Validate a portion of the audio files in the full dataset.

    Validation will warn users if clip durations are inconsistent, as well as raise
    an error if audio sample rates differ between clips in the dataset.

    Parameters
    ----------
    audio_paths : List[Path]
        A list of path objects pointing to audio files in the dataset.
    dataset_type : str
        One of `kws` or `asr`, to toggle relevant warnings/errors.
    """
    clip_durations = []
    sample_rates = []

    with progress.Progress() as p:
        task_id = p.add_task(
            f"Reading {len(audio_paths)} audio samples", total=len(audio_paths)
        )
        with multiprocessing.Pool(processes=multiprocessing.cpu_count()) as pool:
            for result in pool.imap_unordered(audio_map_fn, audio_paths):
                clip_durations.append(result["clip_duration"])
                sample_rates.append(result["sample_rate"])
                p.advance(task_id)

    if dataset_type == "kws" and clip_durations.count(clip_durations[0]) != len(
        clip_durations
    ):
        # this isn't necessarily an error, clips will just be truncated/padded
        # to the desired length
        warnings.warn(
            "Clip durations are inconsistent. They will be cropped "
            "or padded at runtime."
        )
    if sample_rates.count(sample_rates[0]) != len(sample_rates):
        raise ValueError(
            f"Found sample rates in the range ({min(sample_rates)}, "
            f"{max(sample_rates)}) Hz. Audio files require consistent sample "
            "rates before uploading."
        )

    print(f"Successfully validated {len(audio_paths)} audio files.")


def validate_transcriptions(
    transcription_lines: List[str],
    audio_paths: List[Path],
) -> None:
    """
    Validate transcriptions for ASR datasets.

    Each line in a transcription file should have the format:
        `<filename>.<audio_format> <sentence>`
    where the first column is the filename in the same directory
    as the transcription file and the sentence portion is a space delimited
    string of words corresponding to the audio file.

    Validation will throw an error if transcription entries do not contain
    filepaths that point to existing audio files in the dataset.

    Parameters
    ----------
    transcription_lines : List[str]
        A list of strings corresponding to the lines in each of the
        transcription files, with each line prepended with the absolute file path
        to the corresponding transcription file.
    audio_paths : List[Path]
        A list of path objects pointing to audio files in the dataset.
    """
    expected_set = set()
    audio_set = set(audio_paths)
    for line in progress.track(
        transcription_lines,
        description=f"Reading {len(transcription_lines)} transcription entries",
    ):
        expected_set.add(Path(line.split()[0]))

    if not expected_set.issubset(audio_set):
        missing = expected_set - audio_set
        raise FileNotFoundError(
            f"Found reference to {missing.pop()} in transcription, "
            "but file does not exist."
        )

    print(f"Successfully validated {len(transcription_lines)} transcription entries.")


def validate_kws_formatting(audio_paths: List[Path]) -> Dict[str, List[str]]:
    """
    Validate KWS datasets for proper directory structure and audio consistency.

    Parameters
    ----------
    audio_paths : List[Path]
        A list of path objects pointing to all audio files inside the data directory.

    Returns
    -------
    dataset_properties : Dict[str, List[str]]
        Dictionary containing a list of keywords that are inferred from
        the dataset.
    """

    # check validation failure conditions
    if len(audio_paths) == 0:
        raise FileNotFoundError(
            "Found no compatible audio files for kws datasets. "
            "Ensure audio is one of mp3, wav or flac."
        )

    all_splits = set()
    all_keywords = set()
    common_path = os.path.commonpath(audio_paths)
    for fpath in progress.track(
        audio_paths, description="Validating dataset formatting"
    ):
        rel_path = fpath.relative_to(common_path) if len(common_path) > 1 else fpath
        if rel_path.parts[0].startswith(("train", "validation", "test")):
            all_splits.add(infer_split(rel_path))
            if not rel_path.parts[1].startswith("_"):
                all_keywords.add(rel_path.parts[1])
        else:
            all_splits.add(rel_path.parts[0])

    # check split conditions are met
    if not set(("test", "train", "validation")).issubset(all_splits):
        raise FileNotFoundError("Did not find train/validation/test folders in dataset")

    # make sure there are directories in the dataset archive with
    # names corresponding to viable keywords
    if len(all_keywords) == 0:
        raise FileNotFoundError("Did not find any keyword directories.")

    # warn if no background noise directory
    if "background_noise" not in all_splits:
        warnings.warn(
            "No background_noise directory found. "
            "Consider adding for optimal kws performance in NengoEdge."
        )

    # warn about unused directories
    for split in all_splits:
        if split not in ["train", "validation", "test", "background_noise"]:
            warnings.warn(
                f"Found directory {split}, which will be unused by NengoEdge."
            )

    dataset_properties = {"allKeywords": sorted(list(all_keywords))}
    return dataset_properties


def validate_asr_formatting(
    audio_paths: List[Path], transcription_paths: List[Path]
) -> None:
    """
    Validate ASR datasets for proper directory structure, audio consistency, and
    transcription formatting.

    Parameters
    ----------
    audio_paths : List[Path]
        A list of path objects pointing to all audio
        files inside the data directory.
    transcription_paths : List[Path]
        A list of path objects pointing to all transcription
        files inside the data directory.
    """

    if len(audio_paths) == 0:
        raise FileNotFoundError(
            "Found no compatible audio files for asr datasets. "
            "Ensure audio is one of mp3, wav or flac."
        )

    if len(transcription_paths) == 0:
        raise FileNotFoundError(
            "Found no compatible transcription files for asr datasets. "
            "Ensure transcriptions are saved in txt files."
        )

    all_splits = set()
    file_paths = audio_paths + transcription_paths
    common_path = os.path.commonpath(file_paths)
    for fpath in progress.track(
        file_paths, description="Validating dataset formatting"
    ):
        all_splits.add(
            infer_split(
                fpath.relative_to(common_path) if len(common_path) > 1 else fpath
            )
        )

    # check split conditions are met
    for split in ["train", "validation", "test"]:
        if split not in all_splits:
            raise FileNotFoundError(
                "Did not find train/validation/test folders in dataset"
            )


def validate_dataset(
    dataset_type: str, data_directory: Path, validation_samples: Optional[int]
) -> List[Path]:
    """
    Validate a dataset based on a given type and directory for that dataset.

    Parameters
    ----------
    dataset_type : str
        Type of dataset. One of kws or asr.
    data_directory : Path
        The directory in which the dataset resides.
    validation_samples : Optional[int]
        Maximum number of audio files to validate before packaging the dataset.
    """
    # ensure we are only inspecting file paths, not directories
    file_paths = [fpath for fpath in data_directory.rglob("*") if fpath.is_file()]

    if len(file_paths) == 0:
        raise FileNotFoundError(f"Found empty directory {data_directory}")

    audio_paths = [f for f in file_paths if f.suffix in [".mp3", ".wav", ".flac"]]
    if validation_samples is None:
        validation_samples = len(audio_paths)
        if len(audio_paths) > 1e5:
            warnings.warn(
                f"Validating {len(audio_paths)} audio files may take many hours. "
                f"Run with --validation-steps 100000 to choose a random subset."
            )

    assert isinstance(validation_samples, int)
    audio_subset = random.sample(audio_paths, validation_samples)

    if dataset_type == "kws":
        # validate formatting over all paths
        validate_kws_formatting(audio_paths)

        # validate audio over subset if necessary
        validate_audio(audio_subset, dataset_type="kws")

    elif dataset_type == "asr":
        txt_paths = [f for f in file_paths if f.suffix in [".txt"]]
        validate_asr_formatting(audio_paths, txt_paths)
        validate_audio(audio_subset, dataset_type="asr")

        # add absolute file path to transcription entries in order
        # to search for corresponding audio files
        transcription_lines = [
            f"{f.parent}/{line}"
            for f in txt_paths
            for line in f.read_text().splitlines()
        ]
        validate_transcriptions(transcription_lines, audio_paths)

    return file_paths



# pylint: disable=missing-docstring

import json
import shutil
from pathlib import Path
from string import ascii_lowercase
from typing import Any, Dict

import numpy as np
import pytest

from nengo_edge import CoralRunner
from nengo_edge.network_runner import NetworkTokenizer
from nengo_edge.tests.test_saved_model_runner import new_tokenizer
from nengo_edge.version import version


@pytest.fixture(scope="module", name="model_path")
def fixture_model_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    tmp_path = tmp_path_factory.mktemp("sentencepiece")
    _, tokenizer_path = new_tokenizer(tmp_path)
    return tokenizer_path


def test_coral_runner(param_dir: Path) -> None:
    binary_path = param_dir / "model_edgetpu.tflite"
    binary_path.touch()

    with pytest.raises(RuntimeError, match="Cannot connect to address"):
        CoralRunner(directory=param_dir, username="user", hostname="host")

    net_runner = CoralRunner(
        directory=param_dir, username="user", hostname="host", local=True
    )

    assert isinstance(net_runner.remote_dir, Path)
    assert (net_runner.remote_dir / "parameters.json").exists()
    assert (net_runner.remote_dir / "np_mfcc.py").exists()


def prepare_tokenizer(
    temp_dir: Path,
    sp_model_path: Path,
    layer: str = "preprocessing",
    model_type: str = "asr",
) -> None:
    params: Dict[str, Dict[str, Any]] = {
        "preprocessing": {},
        "model": {
            "return_sequences": True,
            "type": model_type,
        },
        "postprocessing": {},
        "version": {"nengo-edge": version},
    }
    if layer in params:
        params[layer]["tokenizer_file"] = sp_model_path.name

    with open(temp_dir / "parameters.json", "w", encoding="utf-8") as f:
        json.dump(params, f)

    shutil.copy(sp_model_path, temp_dir / sp_model_path.name)


@pytest.mark.parametrize("layer", ("preprocessing", "postprocessing"))
def test_paths(layer: str, model_path: Path, tmp_path: Path) -> None:
    prepare_tokenizer(tmp_path, sp_model_path=model_path, model_type="asr", layer=layer)
    network_tokenizer = NetworkTokenizer(tmp_path, "name", "host", local=True)
    assert network_tokenizer.tokenizer_file is not None
    assert network_tokenizer.tokenizer_file == model_path.name
    assert (network_tokenizer.remote_dir / model_path.name).exists()


def test_errors(rng: np.random.RandomState, model_path: Path, tmp_path: Path) -> None:
    # test error with missing tokenizer_file entry in the parameter file
    prepare_tokenizer(tmp_path, sp_model_path=model_path, layer="no layer")
    with pytest.raises(TypeError, match="does not contain any tokenizers"):
        NetworkTokenizer(tmp_path, "name", "host", local=True)

    # test wrong input type on detokenization
    prepare_tokenizer(tmp_path, sp_model_path=model_path)
    network_tokenizer = NetworkTokenizer(tmp_path, "name", "host", local=True)
    with pytest.raises(ValueError, match="must be one of int32/int64"):
        network_tokenizer.run(rng.randint(0, 255, (32, 50)).astype("float32"))

    # test wrong number of dimension on input
    with pytest.raises(ValueError, match="inputs must have exactly 2 dimensions"):
        network_tokenizer.run(rng.randint(0, 255, (32,)).astype("int32"))


def test_tokenize_detokenize(
    rng: np.random.RandomState, param_dir: Path, model_path: Path, tmp_path: Path
) -> None:
    prepare_tokenizer(tmp_path, sp_model_path=model_path)
    network_tokenizer = NetworkTokenizer(tmp_path, "name", "host", local=True)

    input_text = [
        " ".join(
            "".join(rng.choice([*ascii_lowercase], size=3))  # type: ignore
            for _ in range(10)
        )
        for i in range(32)
    ]
    token_ids = network_tokenizer.run(input_text)
    max_len = np.max([len(t) for t in token_ids])
    token_ids = np.array(
        [
            np.concatenate([t, np.full(max_len - len(t), -1, dtype="int32")])
            for t in token_ids
        ]
    )
    decoded_text = network_tokenizer.run(token_ids)

    assert all(t == t_decoded for t, t_decoded in zip(input_text, decoded_text))
